import multiprocessing as mp
import os
from pathlib import Path
import subprocess
import sys

import numpy as np

from libmuscle import Instance, Message
from ymmsl import Operator

from .conftest import skip_if_python_only


def run_macro(instance_id: str, manager_location: str):
    sys.argv.append(f'--muscle-instance={instance_id}')
    sys.argv.append(f'--muscle-manager={manager_location}')
    macro()


def macro():
    instance = Instance({
            Operator.O_I: ['out'],
            Operator.S: ['in']})

    while instance.reuse_instance():
        # f_init
        assert instance.get_setting('test1') == 13

        for i in range(2):
            # o_i
            test_array = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
            assert test_array.shape == (2, 3)
            assert test_array.flags.c_contiguous
            data = {
                    'message': 'testing',
                    'test_grid': test_array}
            instance.send('out', Message(i * 10.0, (i + 1) * 10.0, data))

            # s/b
            msg = instance.receive('in')
            assert msg.data['reply'] == 'testing back {}'.format(i)
            assert msg.data['test_grid'].array.dtype.kind == 'i'
            assert msg.data['test_grid'].array.dtype.itemsize == 8
            assert msg.data['test_grid'].array[0][1] == 2
            assert msg.timestamp == i * 10.0


@skip_if_python_only
def test_cpp_macro_micro(mmp_server_process_simple, tmp_path):
    # create C++ micro model
    # see libmuscle/cpp/src/libmuscle/tests/micro_model_test.cpp
    cpp_build_dir = Path(__file__).parents[1] / 'libmuscle' / 'cpp' / 'build'
    env = os.environ.copy()
    lib_paths = [cpp_build_dir / 'msgpack' / 'msgpack' / 'lib']
    if 'LD_LIBRARY_PATH' in env:
        env['LD_LIBRARY_PATH'] += ':' + ':'.join(map(str, lib_paths))
    else:
        env['LD_LIBRARY_PATH'] = ':'.join(map(str, lib_paths))

    env['MUSCLE_MANAGER'] = mmp_server_process_simple
    cpp_test_dir = cpp_build_dir / 'libmuscle' / 'tests'
    cpp_test_micro = cpp_test_dir / 'micro_model_test'

    with (tmp_path / 'cpp_stdout.txt').open('w') as f_out:
        with (tmp_path / 'cpp_stderr.txt').open('w') as f_err:
            micro_result = subprocess.Popen(
                    [str(cpp_test_micro), '--muscle-instance=micro'], env=env,
                    stdout=f_out, stderr=f_err)

    # run macro model
    macro_process = mp.Process(
            target=run_macro, args=('macro', mmp_server_process_simple))
    macro_process.start()

    # check results
    micro_result.wait()
    assert micro_result.returncode == 0
    macro_process.join()
    assert macro_process.exitcode == 0



import logging
import multiprocessing as mp
import os
from pathlib import Path

import pytest
import yatiml
import ymmsl

import integration_test.include_libmuscle   # noqa: F401

from libmuscle.manager.manager import Manager
from libmuscle.manager.run_dir import RunDir


skip_if_python_only = pytest.mark.skipif(
        'MUSCLE_TEST_PYTHON_ONLY' in os.environ,
        reason='Python-only tests requested')


@pytest.fixture
def yatiml_log_warning():
    yatiml.logger.setLevel(logging.WARNING)


def start_mmp_server(control_pipe, ymmsl_doc, run_dir):
    control_pipe[0].close()
    manager = Manager(ymmsl_doc, run_dir)
    control_pipe[1].send(manager.get_server_location())
    control_pipe[1].recv()
    control_pipe[1].close()
    manager.stop()


def make_server_process(ymmsl_doc, tmpdir):
    run_dir = RunDir(Path(tmpdir))
    control_pipe = mp.Pipe()
    process = mp.Process(target=start_mmp_server,
                         args=(control_pipe, ymmsl_doc, run_dir),
                         name='Manager')
    process.start()
    control_pipe[1].close()
    # wait for start
    yield control_pipe[0].recv()
    control_pipe[0].send(True)
    control_pipe[0].close()
    process.join()


@pytest.fixture
def mmp_server_process(yatiml_log_warning, tmpdir):
    ymmsl_text = (
            'ymmsl_version: v0.1\n'
            'model:\n'
            '  name: test_model\n'
            '  components:\n'
            '    macro: macro_implementation\n'
            '    micro:\n'
            '      implementation: micro_implementation\n'
            '      multiplicity: [10]\n'
            '  conduits:\n'
            '    macro.out: micro.in\n'
            '    micro.out: macro.in\n'
            'settings:\n'
            '  test1: 13\n'
            '  test2: 13.3\n'
            '  test3: testing\n'
            '  test4: True\n'
            '  test5: [2.3, 5.6]\n'
            '  test6:\n'
            '    - [1.0, 2.0]\n'
            '    - [3.0, 1.0]\n'
            'implementations:\n'
            '  macro_implementation: macro.py\n'
            '  micro_implementation: micro.py\n'
            )
    ymmsl_doc = ymmsl.load(ymmsl_text)

    yield from make_server_process(ymmsl_doc, tmpdir)


@pytest.fixture
def mmp_server_process_simple(tmpdir, yatiml_log_warning):
    ymmsl_text = (
            'ymmsl_version: v0.1\n'
            'model:\n'
            '  name: test_model\n'
            '  components:\n'
            '    macro: macro_implementation\n'
            '    micro: micro_implementation\n'
            '  conduits:\n'
            '    macro.out: micro.in\n'
            '    micro.out: macro.in\n'
            'settings:\n'
            '  test1: 13\n'
            '  test2: 13.3\n'
            '  test3: testing\n'
            '  test4: True\n'
            '  test5: [2.3, 5.6]\n'
            '  test6:\n'
            '    - [1.0, 2.0]\n'
            '    - [3.0, 1.0]\n'
            )
    ymmsl_doc = ymmsl.load(ymmsl_text)

    yield from make_server_process(ymmsl_doc, tmpdir)


@pytest.fixture
def mmp_server(yatiml_log_warning):
    ymmsl_text = (
            'ymmsl_version: v0.1\n'
            'model:\n'
            '  name: test_model\n'
            '  components:\n'
            '    macro: macro_implementation\n'
            '    micro:\n'
            '      implementation: micro_implementation\n'
            '      multiplicity: [10]\n'
            '  conduits:\n'
            '    macro.out: micro.in\n'
            '    micro.out: macro.in\n'
            'settings:\n'
            '  test1: 13\n'
            '  test2: 13.3\n'
            '  test3: testing\n'
            '  test4: True\n'
            '  test5: [2.3, 5.6]\n'
            '  test6:\n'
            '    - [1.0, 2.0]\n'
            '    - [3.0, 1.0]\n'
            )
    ymmsl_doc = ymmsl.load(ymmsl_text)

    manager = Manager(ymmsl_doc)
    yield manager._server
    manager.stop()


@pytest.fixture
def log_file_in_tmpdir(tmpdir):
    old_workdir = os.getcwd()
    os.chdir(tmpdir)

    yield None

    os.chdir(old_workdir)



from random import uniform
from time import perf_counter, sleep
from typing import Any, Dict, Iterable, List, Tuple

import msgpack
from ymmsl import Conduit, Operator, Port, Reference, Settings

from libmuscle.mcp.protocol import RequestType, ResponseType
from libmuscle.mcp.tcp_transport_client import TcpTransportClient
from libmuscle.profiling import ProfileEvent
from libmuscle.logging import LogMessage


CONNECTION_TIMEOUT = 300
PEER_TIMEOUT = 600
PEER_INTERVAL_MIN = 5.0
PEER_INTERVAL_MAX = 10.0


def encode_operator(op: Operator) -> str:
    """Convert an Operator to a MsgPack-compatible value."""
    return op.name


def encode_port(port: Port) -> List[str]:
    """Convert a Port to a MsgPack-compatible value."""
    return [str(port.name), encode_operator(port.operator)]


def encode_profile_event(event: ProfileEvent) -> Any:
    """Converts a ProfileEvent to a list.

    Args:
        event: A profile event

    Returns:
        A list with its attributes, for MMP serialisation.
    """
    encoded_port = encode_port(event.port) if event.port else None
    return [
            str(event.instance_id),
            event.start_time.seconds, event.stop_time.seconds,
            event.event_type.value,
            encoded_port, event.port_length, event.slot,
            event.message_size]


class MMPClient():
    """The client for the MUSCLE Manager Protocol.

    This class connects to the Manager and communicates with it on
    behalf of the rest of libmuscle.

    It manages the connection, and converts between our native types
    and the gRPC generated types.
    """
    def __init__(self, location: str) -> None:
        """Create an MMPClient

        Args:
            location: A connection string of the form hostname:port
        """
        self._transport_client = TcpTransportClient(location)

    def close(self) -> None:
        """Close the connection

        This closes the connection. After this no other member
        functions can be called.
        """
        self._transport_client.close()

    def submit_log_message(self, message: LogMessage) -> None:
        """Send a log message to the manager.

        Args:
            message: The message to send.
        """
        request = [
                RequestType.SUBMIT_LOG_MESSAGE.value,
                message.instance_id, message.timestamp.seconds,
                message.level.value, message.text]
        self._call_manager(request)

    def submit_profile_events(self, events: Iterable[ProfileEvent]) -> None:
        """Sends profiling events to the manager.

        Args:
            events: The events to send.
        """
        request = [
                RequestType.SUBMIT_PROFILE_EVENTS.value,
                [encode_profile_event(e) for e in events]]
        self._call_manager(request)

    def get_settings(self) -> Settings:
        """Get the central settings from the manager.

        Returns:
            The requested settings.
        """
        request = [RequestType.GET_SETTINGS.value]
        response = self._call_manager(request)
        return Settings(response[1])

    def register_instance(self, name: Reference, locations: List[str],
                          ports: List[Port]) -> None:
        """Register a component instance with the manager.

        Args:
            name: Name of the instance in the simulation.
            locations: List of places where the instance can be
                    reached.
            ports: List of ports of this instance.
        """
        request = [
                RequestType.REGISTER_INSTANCE.value,
                str(name), locations,
                [encode_port(p) for p in ports]]
        response = self._call_manager(request)
        if len(response) > 1:
            raise RuntimeError(
                    f'Error registering instance: {response[1]}')

    def request_peers(
            self, name: Reference) -> Tuple[
                    List[Conduit],
                    Dict[Reference, List[int]],
                    Dict[Reference, List[str]]]:
        """Request connection information about peers.

        This will repeat the request at an exponentially increasing
        query interval at first, until it reaches the interval
        specified by PEER_INTERVAL_MIN and PEER_INTERVAL_MAX. From
        there on, intervals are drawn randomly from that range.

        Args:
            name: Name of the current instance.

        Returns:
            A tuple containing a list of conduits that this instance is
            attached to, a dictionary of peer dimensions, which is
            indexed by Reference to the peer kernel, and specifies how
            many instances of that kernel there are, and a dictionary
            of peer instance locations, indexed by Reference to a peer
            instance, and containing for each peer instance a list of
            network location strings at which it can be reached.
        """
        sleep_time = 0.1
        start_time = perf_counter()

        request = [RequestType.GET_PEERS.value, str(name)]
        response = self._call_manager(request)

        while (response[0] == ResponseType.PENDING.value and
               perf_counter() < start_time + PEER_TIMEOUT and
               sleep_time < PEER_INTERVAL_MIN):
            sleep(sleep_time)
            response = self._call_manager(request)
            sleep_time *= 1.5

        while (response[0] == ResponseType.PENDING.value and
               perf_counter() < start_time + PEER_TIMEOUT):
            sleep(uniform(PEER_INTERVAL_MIN, PEER_INTERVAL_MAX))
            response = self._call_manager(request)

        if response[0] == ResponseType.PENDING.value:
            raise RuntimeError('Timeout waiting for peers to appear')

        if response[0] == ResponseType.ERROR.value:
            raise RuntimeError('Error getting peers from manager: {}'.format(
                    response[1]))

        conduits = [Conduit(snd, recv) for snd, recv in response[1]]

        peer_dimensions = {
                Reference(component): dims
                for component, dims in response[2].items()}

        peer_locations = {
                Reference(instance): locs
                for instance, locs in response[3].items()}

        return conduits, peer_dimensions, peer_locations

    def deregister_instance(self, name: Reference) -> None:
        """Deregister a component instance with the manager.

        Args:
            name: Name of the instance in the simulation.
        """
        request = [RequestType.DEREGISTER_INSTANCE.value, str(name)]
        response = self._call_manager(request)

        if response[0] == ResponseType.ERROR.value:
            raise RuntimeError('Error deregistering instance: {}'.format(
                    response[1]))

    def _call_manager(self, request: Any) -> Any:
        """Call the manager and do en/decoding.

        Args:
            request: The request to encode and send

        Returns:
            The decoded response
        """
        encoded_request = msgpack.packb(request, use_bin_type=True)
        response = self._transport_client.call(encoded_request)
        return msgpack.unpackb(response, raw=False)



import errno
import logging
from typing import Any, cast, Generator, List

import msgpack
from ymmsl import Conduit, Identifier, Operator, Port, Reference, Settings

from libmuscle.logging import LogLevel
from libmuscle.manager.instance_registry import (
        AlreadyRegistered, InstanceRegistry)
from libmuscle.manager.logger import Logger
from libmuscle.manager.topology_store import TopologyStore
from libmuscle.mcp.protocol import RequestType, ResponseType
from libmuscle.mcp.tcp_transport_server import TcpTransportServer
from libmuscle.mcp.transport_server import RequestHandler
from libmuscle.timestamp import Timestamp
from libmuscle.util import generate_indices, instance_indices


_logger = logging.getLogger(__name__)


def decode_operator(data: str) -> Operator:
    """Create an Operator from a MsgPack-compatible value."""
    return Operator[data]


def decode_port(data: List[str]) -> Port:
    """Create a Port from a MsgPack-compatible value."""
    return Port(Identifier(data[0]), decode_operator(data[1]))


def encode_conduit(conduit: Conduit) -> List[str]:
    """Convert a Conduit to a MsgPack-compatible value."""
    return [str(conduit.sender), str(conduit.receiver)]


class MMPRequestHandler(RequestHandler):
    """Handles Manager requests."""
    def __init__(
            self,
            logger: Logger,
            settings: Settings,
            instance_registry: InstanceRegistry,
            topology_store: TopologyStore):
        """Create an MMPRequestHandler.

        Args:
            logger: The Logger component to log messages to.
            settings: The global settings to serve to instances.
            instance_registry: The database for instances.
            topology_store: Keeps track of how to connect things.
        """
        self._logger = logger
        self._settings = settings
        self._instance_registry = instance_registry
        self._topology_store = topology_store

    def handle_request(self, request: bytes) -> bytes:
        """Handles a manager request.

        Args:
            request: The encoded request

        Returns:
            response: An encoded response
        """
        req_list = msgpack.unpackb(request, raw=False)
        req_type = req_list[0]
        req_args = req_list[1:]
        if req_type == RequestType.REGISTER_INSTANCE.value:
            response = self._register_instance(*req_args)
        elif req_type == RequestType.GET_PEERS.value:
            response = self._get_peers(*req_args)
        elif req_type == RequestType.DEREGISTER_INSTANCE.value:
            response = self._deregister_instance(*req_args)
        elif req_type == RequestType.GET_SETTINGS.value:
            response = self._get_settings(*req_args)
        elif req_type == RequestType.SUBMIT_LOG_MESSAGE.value:
            response = self._submit_log_message(*req_args)
        elif req_type == RequestType.SUBMIT_PROFILE_EVENTS.value:
            response = self._submit_profile_events(*req_args)

        return cast(bytes, msgpack.packb(response, use_bin_type=True))

    def _register_instance(
            self, instance_id: str, locations: List[str],
            ports: List[List[str]]) -> Any:
        """Handle a register instance request.

        Args:
            instance_id: ID of the instance to register
            locations: Locations where it can be reached

        Returns:
            A list containing the following values:

            status (ResponseType): SUCCESS or ERROR
            error_msg (str): An error message, only present if status
                equals ERROR
        """
        port_objs = [decode_port(p) for p in ports]
        try:
            self._instance_registry.add(
                Reference(instance_id), locations, port_objs)

            _logger.info(f'Registered instance {instance_id}')
            return [ResponseType.SUCCESS.value]
        except AlreadyRegistered:
            return [
                    ResponseType.ERROR.value,
                    f'An instance with name {instance_id} was already'
                    ' registered. Did you start a non-MPI component using'
                    ' mpirun?']

    def _get_peers(self, instance_id: str) -> Any:
        """Handle a get peers request.

        Args:
            instance_id: ID of the instance requesting peers

        Returns:
            A list containing the following values on success:

            status (ResponseType): SUCCESS
            conduits (List[List[str]]): Conduits from/to peers
            dimensions (Dict[str, List[int]]): Dimensions of peer
                components
            locations (Dict[str, List[str]]): Locations where peer
                instances can be contacted.

            Or the following values on error:

            status (ResponseType): ERROR
            error_msg (str): An error message

            Or the following values if the result is not yet available:

            status (ResponseType): PENDING
            status_msg (str): A message on what we're waiting for.
        """
        # get info from yMMSL
        instance = Reference(instance_id)
        component = instance.without_trailing_ints()
        if not self._topology_store.has_kernel(component):
            return [ResponseType.ERROR.value, f'Unknown component {component}']

        conduits = self._topology_store.get_conduits(component)
        mmp_conduits = [encode_conduit(c) for c in conduits]

        peer_dims = self._topology_store.get_peer_dimensions(component)
        mmp_dimensions = {str(name): dims for name, dims in peer_dims.items()}

        # generate instances
        try:
            instance_locations = {
                    str(peer): self._instance_registry.get_locations(peer)
                    for peer in self._generate_peer_instances(instance)}
        except KeyError as e:
            return [
                    ResponseType.PENDING.value,
                    f'Waiting for component {e.args[0]}']

        _logger.debug(f'Sent peers to {instance_id}')
        return [
                ResponseType.SUCCESS.value,
                mmp_conduits, mmp_dimensions, instance_locations]

    def _deregister_instance(self, instance_id: str) -> Any:
        """Handle a deregister instance request.

        Args:
            instance_id: ID of the instance to deregister

        Returns:
            A list containing the following values on success:

            status (ResponseType): SUCCESS

            Or the following if an error occurred:

            status (ResponseType): ERROR
            error_msg (str): An error message
        """
        try:
            self._instance_registry.remove(Reference(instance_id))
            _logger.info(f'Deregistered instance {instance_id}')
            return [ResponseType.SUCCESS.value]
        except ValueError:
            return [
                    ResponseType.ERROR.value,
                    f'No instance with name {instance_id} was registered']

    def _get_settings(self) -> Any:
        """Handle a get settings request.

        Returns:
            A list containing the following values on success:

            status (ResponseType): SUCCESS
            settings (Dict[str, SettingValue]): The global settings
        """
        return [
                ResponseType.SUCCESS.value,
                self._settings.as_ordered_dict()]

    def _submit_log_message(
            self, instance_id: str, timestamp: float, level: int, text: str
            ) -> Any:
        """Handle a submit log message request.

        Args:
            instance_id: Sending instance
            timestamp: Time since epoch of the logged event
            level: Log level of the message
            text: Message text

        Returns:
            A list containing the following values on success:

            status (ResponseType): SUCCESS
        """
        self._logger.log_message(
                instance_id, Timestamp(timestamp), LogLevel(level), text)
        return [ResponseType.SUCCESS.value]

    def _submit_profile_events(self, events: List[List[Any]]) -> Any:
        """Handle a submit profile events request.

        Not implemented yet.

        Returns:
            A list containing the following values on success:

            status (ResponseType): SUCCESS
        """
        return [ResponseType.SUCCESS.value]

    def _generate_peer_instances(
            self, instance: Reference) -> Generator[Reference, None, None]:
        """Generates the names of all peer instances of an instance.

        Args:
            instance: The instance whose peers to generate.

        Yields:
            All peer instance identifiers.
        """
        component = instance.without_trailing_ints()
        indices = instance_indices(instance)
        dims = self._topology_store.kernel_dimensions[component]
        all_peer_dims = self._topology_store.get_peer_dimensions(component)
        for peer, peer_dims in all_peer_dims.items():
            base = peer
            for i in range(min(len(dims), len(peer_dims))):
                base += indices[i]

            if dims >= peer_dims:
                yield base
            else:
                for peer_indices in generate_indices(peer_dims[len(dims):]):
                    yield base + peer_indices


class MMPServer:
    """The MUSCLE Manager Protocol server.

    This class accepts connections from the instances comprising
    the multiscale model to be executed, and services them using an
    MMPRequestHandler.
    """
    def __init__(
            self,
            logger: Logger,
            settings: Settings,
            instance_registry: InstanceRegistry,
            topology_store: TopologyStore
            ) -> None:
        """Create an MMPServer.

        This starts a TCP Transport server and connects it to an
        MMPRequestHandler, which uses the given components to service
        the requests. By default, we listen on port 9000, unless it's
        not available in which case we use a random other one.

        Args:
            logger: Logger to send log messages to
            settings: Settings component to get settings from
            instance_registry: To register instances with and get
                peer locations from
            topology_store: To get peers and conduits from
        """
        self._handler = MMPRequestHandler(
                logger, settings, instance_registry, topology_store)
        try:
            self._server = TcpTransportServer(self._handler, 9000)
        except OSError as e:
            if e.errno != errno.EADDRINUSE:
                raise
            self._server = TcpTransportServer(self._handler)

    def get_location(self) -> str:
        """Returns this server's network location.

        This is a string of the form tcp:<hostname>:<port>.
        """
        return self._server.get_location()

    def stop(self) -> None:
        """Stops the server.

        This makes the server stop serving requests, and shuts down its
        background threads.
        """
        self._server.close()



from typing import List, Optional

import numpy as np


class Grid:
    """Represents a grid of data to send or receive.

    Note that for received grids, the array of data is a read-only
    NumPy array. If you have another array that you want to put the
    received data into, use ``np.copyto(dest, source)`` to copy the
    contents of the received array across into your destination array.
    If you don't have an array yet and want a writable version of the
    received array, use ``array.copy()`` to create a writable copy.
    See the tutorial for examples.

    Attributes:
        array (np.ndarray): An array of data
        indexes (Optional[List[str]]): The names of the array's indexes.
    """
    def __init__(
            self, array: np.ndarray, indexes: Optional[List[str]] = None
            ) -> None:
        """Creates a Grid object.

        A Grid object represents an multi-dimensional array of data. It
        has a type, a shape, and optionally a list of index names.

        Supported data types are 4- and 8-byte integers (numpy.int32,
        numpy.int64), 4- and 8-byte floats (numpy.float32,
        numpy.float64), and booleans (np.bool_, np.bool8). The ``data``
        argument must be a NumPy array of one of those types.

        If ``indexes`` is given, then it must be a list of strings of
        the same length as the number of dimensions of ``data``, and
        contain the names of the indexes of the array. For a 2D
        Cartesian grid, these may be ``'x'`` and ``'y'`` for example,
        or for a polar grid, ``'phi'`` and ``'rho'``.

        Args:
            array: An array of data, of a supported type (see above).
            indexes: Names of the indexes (see above).
        """
        if indexes is not None:
            if len(indexes) != array.ndim:
                raise ValueError(
                        'Number of indexes must match number of array'
                        ' dimensions')

        self.array = array
        self.indexes = indexes



import sys
from typing import Sequence

import click
import ymmsl
from ymmsl import Identifier, PartialConfiguration


from libmuscle.planner.planner import Planner, Resources


_RESOURCES_INCOMPLETE_MODEL = """
A model, implementations and resources must be given to be able to calculate
resources for a model. Please ensure that you have specified all of them in
your yMMSL file(s).
"""


@click.group()
def muscle3():
    """MUSCLE3 command line interface

    In the future, this command will provide various functions for
    running coupled simulations using MUSCLE3. For now, it does only
    one thing, which is to calculate the number of cluster nodes
    needed for a given simulation to run without oversubscribing.

    Use muscle3 <command> --help for help with individual commands.
    """
    pass


@muscle3.command(short_help='Calculate resources needed for a simulation')
@click.argument(
        'ymmsl_files', nargs=-1, required=True, type=click.Path(
            exists=True, file_okay=True, dir_okay=False, readable=True,
            allow_dash=True, resolve_path=True))
@click.option(
        '-c', '--cores-per-node', nargs=1, type=int, required=True,
        help='Set number of cores per cluster node.')
@click.option(
        '-v', '--verbose', is_flag=True, help='Show instance allocations.')
def resources(
        ymmsl_files: Sequence[str],
        cores_per_node: int, verbose: bool
        ) -> None:
    """Calculate the number of nodes needed to run the simulation.

    In order to run a MUSCLE3 simulation on a cluster, a batch job has
    to be submitted specifying the resources needed. This command
    calculates the number of nodes needed to run a given simulation
    without oversubscribing, given the number of cores available in
    each node.

    If multiple yMMSL files are given, then they will be combined left
    to right, i.e. if there are conflicting declarations, the one from
    the file last given is used.

    Result:

      With the -v option, the allocation of each instance will be printed
      to stdout. Without it, a single number will be printed, the number
      of nodes needed to run the simulation.

    Examples:

      muscle3 resources --cores-per-node 24 simulation.ymmsl

      muscle3 resources -c 16 model.ymmsl resources.ymmsl

      num_nodes=$(muscle3 resources -c 16 simulation.ymmsl)
      sbatch -N ${num_nodes} simulation.sh

    """
    partial_config = _load_ymmsl_files(ymmsl_files)
    try:
        config = partial_config.as_configuration()
    except ValueError:
        click.echo(_RESOURCES_INCOMPLETE_MODEL, err=True)
        sys.exit(1)

    resources = Resources({'node000001': set(range(cores_per_node))})
    planner = Planner(resources)
    try:
        allocations = planner.allocate_all(config, True)
    except InsufficientResourcesAvailable as e:
        click.echo(e)
        sys.exit(1)

    num_nodes = len(tuple(resources.nodes()))

    if verbose:
        click.echo()
        if num_nodes == 1:
            click.echo('A total of 1 node will be needed, as follows:')
        else:
            click.echo(
                    f'A total of {num_nodes} nodes will be needed,'
                    ' as follows:')

        click.echo()
        for instance in sorted(allocations):
            click.echo(f'{instance}: {str(allocations[instance])}')
    else:
        click.echo(f'{num_nodes}', nl=False)

    sys.exit(0)


def _load_ymmsl_files(ymmsl_files: Sequence[str]) -> PartialConfiguration:
    """Loads and merges yMMSL files."""
    configuration = PartialConfiguration()
    for path in ymmsl_files:
        with open(path, 'r') as f:
            configuration.update(ymmsl.load(f))
    return configuration


if __name__ == '__main__':
    muscle3()



import sys
from typing import Generator
from unittest.mock import MagicMock, patch

import pytest
from ymmsl import Operator, Reference, Settings

from libmuscle.communicator import Message
from libmuscle.instance import Instance
from libmuscle.mpp_message import ClosePort
from libmuscle.settings_manager import SettingsManager


@pytest.fixture
def sys_argv_manager() -> Generator[None, None, None]:
    old_argv = sys.argv
    sys.argv = sys.argv + ['--muscle-manager=localhost:9000']
    yield None
    sys.argv = old_argv


@pytest.fixture
def log_file_in_tmpdir(tmpdir) -> Generator[None, None, None]:
    old_argv = sys.argv
    sys.argv = sys.argv + ['--muscle-log-file={}'.format(tmpdir)]
    yield None
    sys.argv = old_argv


@pytest.fixture
def sys_argv_instance() -> Generator[None, None, None]:
    old_argv = sys.argv
    sys.argv = ['', '--muscle-instance=test_instance[13][42]']
    yield
    sys.argv = old_argv


@pytest.fixture
def instance(sys_argv_instance):
    with patch('libmuscle.instance.MMPClient') as mmp_client, \
         patch('libmuscle.instance.Communicator') as comm_type:
        communicator = MagicMock()
        settings = Settings()
        settings['test1'] = 12
        msg = Message(0.0, 1.0, 'message', settings)
        communicator.receive_message.return_value = msg
        comm_type.return_value = communicator

        mmp_client_object = MagicMock()
        mmp_client_object.request_peers.return_value = (None, None, None)
        mmp_client.return_value = mmp_client_object

        instance = Instance({
            Operator.F_INIT: ['in', 'not_connected'],
            Operator.O_F: ['out']})
        instance._f_init_cache = dict()
        instance._f_init_cache[('in', None)] = msg
        yield instance


@pytest.fixture
def instance2(sys_argv_instance):
    with patch('libmuscle.instance.MMPClient') as mmp_client, \
         patch('libmuscle.instance.Communicator'):
        mmp_client_object = MagicMock()
        mmp_client_object.request_peers.return_value = (None, None, None)
        mmp_client.return_value = mmp_client_object
        instance = Instance({
            Operator.F_INIT: ['in[]'],
            Operator.O_F: ['out']})
        yield instance


def test_create_instance(
        sys_argv_instance, log_file_in_tmpdir, sys_argv_manager):
    with patch('libmuscle.instance.MMPClient') as mmp_client, \
         patch('libmuscle.instance.Communicator') as comm_type:
        mmp_client_object = MagicMock()
        mmp_client_object.request_peers.return_value = (None, None, None)
        mmp_client.return_value = mmp_client_object
        ports = {
            Operator.F_INIT: ['in'],
            Operator.O_F: ['out']}
        instance = Instance(ports)
        assert instance._name == Reference('test_instance')
        assert instance._index == [13, 42]
        assert instance._declared_ports == ports
        assert isinstance(instance._settings_manager, SettingsManager)
        assert len(instance._settings_manager.base) == 0
        assert len(instance._settings_manager.overlay) == 0
        mmp_client.assert_called_once_with('localhost:9000')
        assert mmp_client_object._register.called_with()
        assert mmp_client_object._connect.called_with()
        comm_type.assert_called_with(Reference('test_instance'), [13, 42],
                                     ports, instance._profiler)
        assert instance._communicator == comm_type.return_value
        assert isinstance(instance._settings_manager, SettingsManager)
        assert len(instance._settings_manager.base) == 0


def test_extract_manager_location(sys_argv_manager) -> None:
    assert (Instance._Instance__extract_manager_location() ==
            'localhost:9000')


def test_get_setting(instance):
    ref = Reference
    settings = Settings()
    settings[ref('test1')] = 'test'
    settings[ref('test2')] = 12
    settings[ref('test3')] = 27.1
    settings[ref('test4')] = True
    settings[ref('test5')] = [2.3, 5.6]
    settings[ref('test6')] = [[1.0, 2.0], [3.0, 4.0]]
    instance._settings_manager.base = settings

    assert instance.get_setting('test1') == 'test'
    assert instance.get_setting('test2') == 12
    assert instance.get_setting('test3') == 27.1
    assert instance.get_setting('test4') is True
    assert instance.get_setting('test5') == [2.3, 5.6]
    assert instance.get_setting('test6') == [
            [1.0, 2.0], [3.0, 4.0]]

    assert instance.get_setting('test1', 'str') == 'test'
    assert instance.get_setting('test2', 'int') == 12
    assert instance.get_setting('test3', 'float') == 27.1
    assert instance.get_setting('test4', 'bool') is True
    assert (instance.get_setting('test5', '[float]') ==
            [2.3, 5.6])
    assert (instance.get_setting('test6', '[[float]]') ==
            [[1.0, 2.0], [3.0, 4.0]])

    with pytest.raises(KeyError):
        instance.get_setting('testx')

    with pytest.raises(TypeError):
        instance.get_setting('test1', 'int')
    with pytest.raises(TypeError):
        instance.get_setting('test6', '[float]')
    with pytest.raises(TypeError):
        instance.get_setting('test5', '[[float]]')
    with pytest.raises(ValueError):
        instance.get_setting('test2', 'nonexistenttype')


def test_list_ports(instance):
    ports = instance.list_ports()
    assert instance._communicator.list_ports.called_with()
    assert ports == instance._communicator.list_ports.return_value


def test_is_vector_port(instance):
    instance._communicator.get_port.return_value.is_vector = MagicMock(
            return_value=True)
    is_vector = instance.is_vector_port('out_port')
    assert is_vector is True
    assert instance._communicator.get_port.called_with('out_port')


def test_send(instance, message):
    instance.send('out', message, 1)
    assert instance._communicator.send_message.called_with(
            'out', message, 1)


def test_send_invalid_port(instance, message):
    instance._communicator.port_exists.return_value = False
    with pytest.raises(RuntimeError):
        instance.send('does_not_exist', message, 1)


def test_receive(instance):
    instance._communicator.get_port.return_value = MagicMock(
            operator=Operator.F_INIT)
    msg = instance.receive('in')
    assert msg.timestamp == 0.0
    assert msg.next_timestamp == 1.0
    assert instance._communicator.receive_message.called_with(
            'in', None)
    assert msg.data == 'message'

    with pytest.raises(RuntimeError):
        instance.receive('in')


def test_receive_default(instance):
    instance._communicator.port_exists.return_value = True
    port = instance._communicator.get_port.return_value
    port.operator = Operator.F_INIT
    port.is_connected.return_value = False
    instance.receive('not_connected', 1, 'testing')
    assert instance._communicator.receive_message.called_with(
            'not_connected', 1, 'testing')
    with pytest.raises(RuntimeError):
        instance.receive('not_connected', 1)


def test_receive_invalid_port(instance):
    instance._communicator.port_exists.return_value = False
    with pytest.raises(RuntimeError):
        instance.receive('does_not_exist', 1)


def test_receive_with_settings(instance):
    msg = instance.receive_with_settings('in', 1)
    assert (instance._communicator.receive_message
            .called_with('in', 1))
    assert msg.timestamp == 0.0
    assert msg.next_timestamp == 1.0
    assert msg.data == 'message'
    assert msg.settings['test1'] == 12


def test_receive_with_settings_default(instance):
    instance.receive_with_settings('not_connected', 1, 'testing')
    assert instance._communicator.receive_message.called_with(
            'not_connected', 1, 'testing')


def test_receive_parallel_universe(instance) -> None:
    instance._settings_manager.overlay['test2'] = 'test'
    with pytest.raises(RuntimeError):
        instance.receive('in')


def test_reuse_instance_receive_overlay(instance):
    instance._settings_manager.overlay = Settings()
    test_base_settings = Settings()
    test_base_settings['test1'] = 24
    test_base_settings['test2'] = [1.3, 2.0]
    test_overlay = Settings()
    test_overlay['test2'] = 'abc'
    recv = instance._communicator.receive_message
    recv.return_value = Message(0.0, None, test_overlay, test_base_settings)
    instance.reuse_instance()
    assert instance._communicator.receive_message.called_with(
        'muscle_settings_in')
    assert len(instance._settings_manager.overlay) == 2
    assert instance._settings_manager.overlay['test1'] == 24
    assert instance._settings_manager.overlay['test2'] == 'abc'


def test_reuse_instance_closed_port(instance):
    def receive_message(port_name, slot=None, default=None):
        if port_name == 'muscle_settings_in':
            return Message(0.0, None, Settings(), Settings())
        elif port_name == 'in':
            return Message(0.0, None, ClosePort(), Settings())
        assert False    # pragma: no cover

    def get_port(port_name):
        port = MagicMock()
        port.is_vector.return_value = False
        if port_name == 'not_connected':
            port.is_connected.return_value = False
        else:
            port.is_connected.return_value = True
        return port

    instance._communicator.receive_message = receive_message
    instance._communicator.list_ports.return_value = {
            Operator.F_INIT: ['in', 'not_connected'],
            Operator.O_F: ['out']}
    instance._communicator.get_port = get_port
    instance._Instance__close_ports = MagicMock()

    do_reuse = instance.reuse_instance()
    assert do_reuse is False


def test_reuse_instance_vector_port(instance2):
    def receive_message(port_name, slot=None, default=None):
        if port_name == 'muscle_settings_in':
            return Message(0.0, None, Settings(), Settings())
        elif port_name == 'in':
            data = 'test {}'.format(slot)
            return Message(0.0, None, data, Settings())
        assert False    # pragma: no cover

    instance2._communicator.receive_message = receive_message
    instance2._communicator.list_ports.return_value = {
            Operator.F_INIT: ['in'],
            Operator.O_F: ['out']}

    port = MagicMock()
    port.is_vector.return_value = True
    port.is_connected.return_value = True
    port.get_length.return_value = 10
    instance2._communicator.get_port.return_value = port

    do_reuse = instance2.reuse_instance()
    assert do_reuse is True

    msg = instance2.receive('in', 5)
    assert msg.timestamp == 0.0
    assert msg.next_timestamp is None
    assert msg.data == 'test 5'


def test_reuse_instance_no_f_init_ports(instance):
    instance._communicator.receive_message.return_value = Message(
            0.0, None, Settings(), Settings())
    instance._communicator.list_ports.return_value = {}
    instance._communicator.settings_in_connected.return_value = False
    do_reuse = instance.reuse_instance()
    assert do_reuse is True
    do_reuse = instance.reuse_instance()
    assert do_reuse is False


def test_reuse_instance_miswired(instance):
    with pytest.raises(RuntimeError):
        instance.reuse_instance()
